# -*- coding: utf-8 -*-

import json

from vaporize.core import convert_datetime, get_url, handle_request, query
from vaporize.utils import DotDict

BACKUP_WEEKLY_DISABLED  = 'DISABLED'
BACKUP_WEEKLY_SUNDAY    = 'SUNDAY'
BACKUP_WEEKLY_MONDAY    = 'MONDAY'
BACKUP_WEEKLY_TUESDAY   = 'TUESDAY'
BACKUP_WEEKLY_WEDNESDAY = 'WEDNESDAY'
BACKUP_WEEKLY_THURSDAY  = 'THURSDAY'
BACKUP_WEEKLY_FRIDAY    = 'FRIDAY'
BACKUP_WEEKLY_SATURDAY  = 'SATURDAY'

BACKUP_DAILY_DISABLED    = 'DISABLED'
BACKUP_DAILY_H_0000_0200 = 'H_0000_0200'
BACKUP_DAILY_H_0200_0400 = 'H_0200_0400'
BACKUP_DAILY_H_0400_0600 = 'H_0400_0600'
BACKUP_DAILY_H_0600_0800 = 'H_0600_0800'
BACKUP_DAILY_H_0800_1000 = 'H_0800_1000'
BACKUP_DAILY_H_1000_1200 = 'H_1000_1200'
BACKUP_DAILY_H_1200_1400 = 'H_1200_1400'
BACKUP_DAILY_H_1400_1600 = 'H_1400_1600'
BACKUP_DAILY_H_1600_1800 = 'H_1600_1800'
BACKUP_DAILY_H_1800_2000 = 'H_1800_2000'
BACKUP_DAILY_H_2000_2200 = 'H_2000_2200'
BACKUP_DAILY_H_2200_0000 = 'H_2200_0000'


class BackupSchedule(DotDict):
    """A CloudServers Backup Schedule."""
    def __repr__(self):
        if 'daily' in self:
            return '<BackupSchedule %s>' % self['daily']
        if 'weekly' in self:
            return '<BackupSchedule %s>' % self['weekly']
        return super(BackupSchedule, self).__repr__()

    @classmethod
    def create(cls, weekly=None, daily=None):
        return cls(weekly=weekly, daily=daily)

    def to_dict(self):
        ret = {'backupSchedule': {'enabled': True}}
        if self.weekly:
            ret['backupSchedule']['weekly'] = self.weekly
        if self.daily:
            self['backupSchedule']['daily'] = daily
        return ret


class Flavor(DotDict):
    """A CloudServers Flavor."""
    def __repr__(self):
        if 'name' in self:
            return '<Flavor %s>' % self['name']
        return super(Flavor, self).__repr__()

    @classmethod
    def list(cls, limit=None, offset=None, detail=False):
        """Returns a list of Flavors.

        :param limit: Limit the result set by a number
        :type limit: int
        :param offset: Offset the result set by a number
        :type offset: int
        :param detail: Return additional details about each Flavor
        :type: bool
        :returns: A list of CloudServers Flavors.
        :rtype: :class:`Flavor`

        .. versionadded:: 0.1
        """
        url = [get_url('cloudservers'), 'flavors']
        if detail:
            url.append('detail')
        url = '/'.join(url)
        if limit is not None or offset is not None:
            url = query(url, limit=limit, offset=offset)
        return handle_request('get', url, wrapper=cls, container='flavors')

    @classmethod
    def get(cls, id):
        """Returns a Flavor by ID.

        :param id: The ID of the Flavor to retrieve
        :type id: int
        :returns: A CloudServers Flavor matching the ID.
        :rtype: :class:`Flavor`

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('cloudservers'), 'flavors', str(id)])
        return handle_request('get', url, wrapper=cls, container='flavor')


class Image(DotDict):
    """A CloudServers Image."""
    def __repr__(self):
        if 'name' in self:
            return '<Image %s>' % self['name']
        return super(Image, self).__repr__()

    def __setitem__(self, key, value):
        if key == 'serverId':
            key = 'server_id'
        elif key in ['created', 'updated']:
            value = convert_datetime(value)
        super(Image, self).__setitem__(key, value)

    def reload(self):
        """Reload this Image (an implicit :func:`get`).

        :returns: An updated CloudServers Image.
        :rtype: :class:`Image`

        .. versionadded:: 0.1.9
        """
        assert 'id' in self, "Missing Image ID"
        response = Image.get(self['id'])
        self.update(response)
        return self

    def delete(self):
        """Delete this Image.

        .. note::

            You can only delete Images you create.

        .. warning::

            There is not confirmation step for this operation. Deleting an
            image is permanent.

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('cloudservers'), 'images', str(self['id'])])
        handle_request('delete', url)

    @classmethod
    def list(cls, limit=None, offset=None, detail=False):
        """Returns a list of CloudServers Images.

        :param limit: Limit the result set by a cetain number
        :type limit: int
        :param offset: Offset the result set by a certain number
        :type offset: int
        :param detail: Return additional details about each Image
        :type detail: bool
        :returns: A list of CloudServers Images.
        :rtype: A list of :class:`Image`

        .. versionadded:: 0.1
        """
        url = [get_url('cloudservers'), 'images']
        if detail:
            url.append('detail')
        url = '/'.join(url)
        if limit is not None or offset is not None:
            url = query(url, limit=limit, offset=offset)
        return handle_request('get', url, wrapper=Image, container='images')

    @classmethod
    def get(cls, id):
        """Return an Image by ID.

        :param id: The ID of the Image to retrieve
        :type id: int
        :returns: A CloudServer Image matching the ID.
        :rtype: :class:`Image`

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('cloudservers'), 'images', str(id)])
        return handle_request('get', url, wrapper=cls, container='image')

    @classmethod
    def create(cls, name, server):
        """Create an Image.

        :param name: Name of the Image
        :type name: str
        :param server: Server or ``id`` to base the Image upon
        :type server: int or :class:`Server`
        :returns: A shiny new CloudServers Image.
        :rtype: :class:`Image`

        .. versionadded:: 0.1
        """
        if isinstance(server, Server):
            server = server.id
        server = int(server)
        data = {'image': {'serverId': server,
                          'name': name}}
        data = json.dumps(data)
        url = '/'.join([get_url('cloudservers'), 'images'])
        return handle_request('post', url, data, cls, 'image')


class IP(DotDict):
    """A CloudServers IP Address."""
    def __repr__(self):
        if 'public' in self:
            return '<IP %s>' % self['public'][0]
        if 'private' in self:
            return '<IP %s>' % self['private'][0]
        return super(IP, self).__repr__()


class Server(DotDict):
    """A CloudServers Server."""
    def __repr__(self):
        if 'name' in self:
            return "<Server %s>" % self['name']
        return super(Server, self).__repr__()

    def __setitem__(self, key, value):
        if key == 'addresses':
            value = IP(value)
        super(Server, self).__setitem__(key, value)

    def reload(self):
        """Reload this Server (an implicit :func:`get`).

        :returns: An updated CloudServers Server.
        :rtype: :class:`Server`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        response = Server.get(self['id'])
        self.update(response)
        return self

    def modify(self, name=None, password=None):
        """Modify this Server's name or root password.

        :param name: Change the Server's name
        :type name: str
        :param password: Change the Server's root password
        :type password: str
        :returns: A modified CloudServers Server.
        :rtype: :class:`Server`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        data = {'server': {}}
        if name is not None:
            data['server']['name'] = name
        if password is not None:
            data['server']['adminPass'] = password
        data = json.dumps(data)
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id'])])
        response = handle_request('put', url, data=data)
        if response:
            if name is not None:
                self['name'] = name
        return self

    def delete(self):
        """Delete this Server.

        .. warning::

            There is no confirmation step for this operation. When you delete a
            server it is permanent. If in doubt, create a backup image
            (:func:`vaporize.images.create`) first before deleting.

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id'])])
        handle_request('delete', url)

    @property
    def ips(self):
        """
        Returns a list of public and private IPs for this Server.

        :returns: A list of public and private IPs for this Server.
        :rtype: A list of :class:`IP`

        .. versionadded:: 0.1
        """
        if 'addresses' not in self:
            assert 'id' in self
            url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                            'ips'])
            response = handle_request('get', url, wrapper=IP,
                                      container='addresses')
            self['addresses'] = response
        return self['addresses']

    @property
    def public_ips(self):
        """Returns the Server's Public IP.

        .. versionadded:: 0.1
        """
        if 'addresses' not in self:
            self['addresses'] = IP()
        if 'public' not in self['addresses']:
            assert 'id' in self
            url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                            'ips', 'public'])
            response = handle_request('get', url, wrapper=IP)
            self['addresses'].update(response)
        return self['addresses']['public']

    @property
    def private_ips(self):
        """Returns the Server's Private IP.

        .. versionadded:: 0.1
        """
        if 'addresses' not in self:
            self['addresses'] = IP()
        if 'private' not in self['addresses']:
            assert 'id' in self
            url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                            'ips', 'private'])
            response = handle_request('get', url, wrapper=IP)
            self['addresses'].update(response)
        return self['addresses']['private']

    def share_ip(self, address, ipgroup, configure=True):
        """Share this Server's IP in a Shared IP Group.

        :param address: IP to share in the Shared IP Group
        :type address: str
        :param ipgroup: A :class:`SharedIPGroup` or ``id``
        :type ipgroup: int or :class:`SharedIPGroup`
        :param configure: Configure the shared IP on the Server
        :type configure: bool
        :returns: The Shared IP Group associated with this Server.
        :rtype: :class:`SharedIPGroup`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        if isinstance(ipgroup, SharedIPGroup):
            ipgroup = ipgroup.id
        ipgroup = int(ipgroup)
        data = json.dumps({'shareIp': {'sharedIpGroup': ipgroup,
                                       'configureServer': configure}})
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                        'ips', 'public', address])
        handle_request('put', url, data=data)

    def unshare_ip(self, address):
        """Unshare this Server's IP

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                        'ips', 'public', address])
        handle_request('delete', url)

    def reboot(self, type='SOFT'):
        """Perform a soft/hard reboot on this Server.

        :param type: A reboot type (``SOFT`` or ``HARD``).
        :type type: str

        .. versionadded:: 0.1
        """
        assert 'id' in self, "Missing Server ID"
        assert type in ['SOFT', 'HARD'], "Reboot type must be 'SOFT' or 'HARD'"
        data = json.dumps({'reboot': {'type': type}})
        url = '/'.join([get_url('cloudservers'), 'servers',
                        str(self['id']), 'action'])
        handle_request('post', url, data)

    def rebuild(self, image):
        """Rebuild this Server using a specified Image

        :param image: The Image or ``id``
        :type image: int or :class:`Image`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        if isinstance(image, Image):
            image = image.id
        image = int(image)
        data = json.dumps({'rebuild': {'imageId': int(image)}})
        url = '/'.join([get_url('cloudservers'), 'servers',
                        str(self['id']), 'action'])
        handle_request('post', url, data)

    def resize(self, flavor):
        """Resize this Server to a specific Flavor size

        :param flavor: The Flavor or ``id``
        :type flavor: int or :class:`Flavor`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        if isinstance(flavor, Flavor):
            flavor = flavor.id
        flavor = int(flavor)
        data = json.dumps({'resize': {'flavorId': flavor}})
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                        'action'])
        handle_request('post', url, data)

    def confirm_resize(self):
        """Confirm a successful resize operation

        .. versionadded:: 0.1
        """
        assert 'id' in self
        data = json.dumps({'confirmResize': None})
        url = '/'.join([get_url('cloudservers'), 'servers',
                        str(self['id']), 'action'])
        handle_request('post', url, data)

    def revert_resize(self):
        """Revert an unsuccessful resize operation

        .. versionadded:: 0.1
        """
        assert 'id' in self
        data = json.dumps({'revertResize': None})
        url = '/'.join([get_url('cloudservers'), 'servers',
                        str(self['id']), 'action'])
        handle_request('post', url, data)

    @property
    def backup_schedule(self):
        """Return this Server's backup schedule

        :return: :class:`BackupSchedule`

        .. versionadded:: 0.1
        """
        if 'backup_schedule' not in self:
            assert 'id' in self
            url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                            'backup_schedule'])
            response = handle_request('get', url, wrapper=BackupSchedule,
                                      container='backupSchedule')
            self['backup_schedule'] = response
        return self['backup_schedule']

    @backup_schedule.setter
    def backup_schedule(self, schedule):
        """Enable a backup schedule for this Server

        :param schedule: A BackupSchedule instance 
        :type schedule: :class:`BackupSchedule`

            >>> server = vaporize.servers.Server.get(...)
            >>> bs = vaporize.servers.BackupSchedule.create(weekly=...,
            ...                                             daily=...)
            >>> server.backup_schedule = bs


        .. versionadded:: 0.1
        """
        assert 'id' in self
        assert isinstance(schedule, BackupSchedule)
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                        'backup_schedule'])
        handle_request('post', url, schedule.to_dict())
        self['backup_schedule'] = schedule

    @backup_schedule.deleter
    def backup_schedule(self):
        """Disable a backup schedule for this Server

            >>> server = vaporize.servers.Server.get(...)
            >>> del server.backup_schedule

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('cloudservers'), 'servers', str(self['id']),
                        'backup_schedule'])
        handle_request('delete', url)
        del self['backup_schedule']

    @classmethod
    def list(cls, limit=None, offset=None, detail=False):
        """
        List of CloudServer Servers

        :param limit: Limit the result set to a certain number
        :type limit: int
        :param offset: Offset the result set by a certain number
        :type offset: int
        :param detail: Return detailed information about each Server
        :type detail: bool
        :returns: A list of CloudServers Servers.
        :rtype: List of :class:`Server`

        .. versionadded:: 0.1
        """
        url = [get_url('cloudservers'), 'servers']
        if detail:
            url.append('detail')
        url = '/'.join(url)
        if limit is not None or offset is not None:
            url = query(url, limit=limit, offset=offset)
        return handle_request('get', url, wrapper=cls, container='servers')

    @classmethod
    def get(cls, id):
        """Return a Server using an ID

        :param id: The ``id`` of the Server to be retrieved
        :type id: int
        :return: A :class:`Server`

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('cloudservers'), 'servers', str(id)])
        return handle_request('get', url, wrapper=cls, container='server')

    @classmethod
    def create(cls, name, image, flavor, metadata=None, files=None):
        """Create a CloudServers Server

        :param name: A Server's name
        :type name: str
        :param image: An Image or ``id``
        :type image: int or :class:`Image`
        :param flavor: A Flavor or ``id``
        :type flavor: int or :class:`Flavor`
        :param metadata: Optional meta data to include with Server
        :type metadata: dict
        :param files: A list of files to load on Server
        :type files: dict
        :returns: A shiny new CloudServers Server.
        :rtype: :class:`Server`

        .. versionadded:: 0.1
        """
        if isinstance(image, Image):
            image = image.id
        image = int(image)
        if isinstance(flavor, Flavor):
            flavor = flavor.id
        flavor = int(flavor)
        data = {'server': {'name': name,
                           'imageId': image,
                           'flavorId': flavor,
                           'metadata': metadata or {},
                           'personality': []}}
        if isinstance(files, dict):
            for path, contents in list(files.items()):
                data['personality'].append({'path': path, 'contents': contents})
        data = json.dumps(data)
        url = '/'.join([get_url('cloudservers'), 'servers'])
        return handle_request('post', url, data, cls, 'server')


class SharedIPGroup(DotDict):
    """A Cloudservers Shared IP Group."""
    def __repr__(self):
        if 'name' in self:
            return '<SharedIPGroup %s>' % self['name']
        return super(SharedIPGroup, self).__repr__()

    def __setitem__(self, key, value):
        if key == 'sharedIpGroupId':
            key = 'id'
        elif key == 'configuredServer':
            key = 'configured'
        super(SharedIPGroup, self).__setitem__(key, value)

    def delete(self):
        """Delete this Shared IP Group.

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('cloudservers'), 'shared_ip_groups',
                        str(self['id'])])
        handle_request('delete', url)

    @classmethod
    def list(cls, limit=None, offset=None, detail=False):
        """Returns a list of Shared IP Groups.

        :param limit: Limit the result set by a certain number
        :type limit: int
        :param offset: Offset the result set by a certain number
        :type offset: int
        :param detail: Return additional details about each Shared IP Group
        :type detail: bool
        :returns: A list of Shared IP Groups.
        :rtype: A list of :class:`SharedIPGroup`

        .. versionadded:: 0.1
        """
        url = [get_url('cloudservers'), 'shared_ip_groups']
        if detail:
            url.append('detail')
        url = '/'.join(url)
        if limit is not None or offset is not None:
            url = query(url, limit=limit, offset=offset)
        return handle_request('get', url, wrapper=cls,
                              container='sharedIpGroups')

    @classmethod
    def get(cls, id):
        """Return a Shared IP Group by ID.

        :param id: The ID of the Shared IP Group to retrieve
        :type id: int
        :returns: A :class:`SharedIPGroup`

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('cloudservers'), 'shared_ip_groups', str(id)])
        return handle_request('get', url, wrapper=cls,
                              container='sharedIpGroup')

    @classmethod
    def create(cls, name, server):
        """Create a Shared IP Group.

        :param name: Name of the Shared IP Group
        :type name: str
        :param server: The Server or ``id`` to add to group
        :type server: int or :class:`Server`
        :returns: A shiny new CloudServers Shared IP Group.
        :rtype: :class:`SharedIPGroup`

        .. versionadded:: 0.1
        """
        if isinstance(server, Server):
            server = server.id
        server = int(server)
        data = {'sharedIpGroup': {'name': name,
                                  'server': server}}
        data = json.dumps(data)
        url = '/'.join([get_url('cloudservers'), 'server_ip_groups'])
        return handle_request('post', url, data, cls, 'sharedIpGroup')



# -*- coding: utf-8 -*-

__title__ = 'vaporize'
__author__ = 'Michael Lavers'
__email__ = 'kolanos@gmail.com'
__version__ = '0.2.1'
__license__ = 'MIT'
__copyright__ = 'Copyright 2012 Michael Lavers'

from . import databases, domains, loadbalancers, servers
from .core import connect



# -*- coding: utf-8 -*-

import re


class DotDict(dict):
    """
    Give dict support for dot notation and force value changes through
    __setitem__ so that it can be overloaded/overidden.
    """
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))
        other = dict(*args, **kwargs)
        for key in other:
            self[key] = other[key]

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]

    def __setitem__(self, key,value):
        key = camelcase_to_underscore(key)
        super(DotDict, self).__setitem__(key, value)

    __setattr__ = __setitem__
    __getattr__ = dict.__getitem__
    __delattr__ = dict.__delitem__

camelcase_to_underscore = lambda str: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', str).lower().strip('_')



# -*- coding: utf-8 -*-

import json

from vaporize.core import convert_datetime, get_url, handle_request, query
from vaporize.utils import DotDict


class Change(DotDict):
    """A CloudDNS Change History."""
    pass


class Domain(DotDict):
    """A CloudDNS Domain."""
    def __repr__(self):
        if 'name' in self:
            return '<Domain %s>' % self.name
        return super(Domain, self).__repr__()

    def __setitem__(self, key, value):
        if key == 'recordsList':
            key = 'records'
            value = [Record(v) for v in value['records']]
        elif key == 'subdomains':
            value = [Subdomain(v) for v in value['domains']]
        elif key in ['created', 'updated']:
            value = convert_datetime(value)
        super(Domain, self).__setitem__(key, value)

    def reload(self):
        """
        Reload this Domain (an implicit :func:`get`).

        :returns: An updated Domain.
        :rtype: :class:`Domain`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        response = get(self['id'])
        self.update(response)
        return self

    def modify(self, ttl=None, email_address=None, comment=None):
        """Modify this Domain's properties.

        :param ttl: Time-To-Live (TTL0 in seconds
        :type ttl: int
        :param email_address: E-mail address associated with Domain
        :type email_address: str
        :param comment: A comment or note about this Domain
        :type comment: str
        :returns: A :class:`Domain`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        data = {}
        if ttl is not None:
            data['ttl'] = int(ttl)
        if email_address is not None:
            data['emailAddress'] = email_address
        if comment is not None:
            data['comment'] = comment
        data = json.dumps(data)
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id'])])
        handle_request('put', url, data)
        if ttl is not None:
            self['ttl'] = int(ttl)
        if email_address is not None:
            self['email_address'] = email_address
        if comment is not None:
            self['comment'] = comment
        return self

    def delete(self, subdomains=False):
        """Delete this Domain.

        .. warning::

            There is no confirmation step to this operation. Deleting this
            domain is permanent. If in doubt you can export a copy of the DNS
            zone (:func:`vaporize.domains.Domain.export_zone`) before deleting.

        :param subdomains: Delete this Domain's Subdomains (optional)
        :type subdomains: bool

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id'])])
        url = query(url, deleteSubdomains=subdomains)
        handle_request('delete', url)

    @property
    def records(self):
        """Returns a list of CloudDNS Records.

        :returns: A list of Records.
        :rtype: A list of :class:`Record`

        .. versionadded:: 0.1
        """
        if 'records' not in self:
            assert 'id' in self
            url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                            'records'])
            response = handle_request('get', url, wrapper=Record,
                                      container='records',
                                      domain_id=self['id'])
            self['records'] = response
        return self['records']

    def add_records(self, *records):
        """Add Records to a Domain.

            >>> domain = vaporize.domains.create(...)
            >>> record1 = vaporize.domains.Record.create(....)
            >>> record2 = vaporize.domains.Record.create(...)
            >>> domain.add_recrods(record1, record2)

        :param records: Records you wish to add to this Domain.
        :type records: :class:`Record`
        :returns: A list of Records
        :rtype: :class:`Record`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        data = {'records': []}
        for record in records:
            if isinstance(record, Record):
                data['records'].append({
                    'name': record.name,
                    'type': record.type,
                    'data': record.data,
                    'ttl': record.ttl,
                    'priority': record.priority,
                    'comment': record.comment
                    })
        data = json.dumps(data)
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                        'records'])
        self['records'] = handle_request('post', url, data, Record, 'records',
                                          domain_id=self['id'])
        return self['records']

    def remove_record(self, record):
        """Remove a Record from this Domain.

        :param record: A Record or ``id`` for the Record to remove.
        :type record: int or :class:`REcord`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        record = record.id if isinstance(record, Record) else int(record)
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                        'records', str(record)])
        handle_request('delete', url)

    @property
    def subdomains(self):
        """Returns a list of Subdomains.

        :returns; A list of Subdomains.
        :rtype: A list of :class:`Subdomain`

        .. versionadded:: 0.1
        """
        if 'subdomins' not in self:
            assert 'id' in self
            url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                            'subdomains'])
            response = handle_request('get', url, wrapper=Subdomain,
                                      container='domains',
                                      domain_id=self['id'])
            self['subdomains'] = response
        return self['subdomains']

    def changes(self, since):
        """Returns a list of CloudDNS changes for this domain.

        :param since: A datetime as a starting point.
        :type since: str or datetime
        :returns: A list of Changes
        :rtype: A list of :class:`Change`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                        'changes'])
        url = query(url, since=str(since))
        return handle_request('get', url, wrapper=Change, container='changes')

    @property
    def export_zone(self):
        """Export the raw BIND zone for this Domain.

        :returns: An :class:`Export` containing the raw BIND zone

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('clouddns'), 'domains', str(self['id']),
                        'export'])
        return handle_request('get', url, wrapper=Export)

    @classmethod
    def list(cls, limit=None, offset=None, filter=None):
        """List of Domains.

        :param limit: Limit the number of results returned
        :type limit: int
        :param offset: Offset the result set by a certain amount
        :type offset: int
        :param filter: Filter results by a domain name
        :type filter: str
        :returns: A list of Domains
        :rtype: A list of :class:`Domain`

        .. versionadded:: 0.1
        """
        url = [get_url('clouddns'), 'domains']
        url = '/'.join(url)
        if limit is not None or offset is not None:
            url = query(url, limit=limit, offset=offset)
        if filter is not None:
            url = query(url, name=filter)
        return handle_request('get', url, wrapper=cls, container='domains')

    @classmethod
    def get(cls, id, records=False, subdomains=False):
        """Retrieve a Domain using an ID.

        :param records: Include the Domain's Records in the result
        :type records: bool
        :param subdomains: Include the Domain's Subdomainsi n the result
        :type subdomains: bool
        :returns: A Domain with the specified ID.
        :rtype: :class:`Domain`

        .. versionadded:: 0.1
        """
        url = '/'.join([get_url('clouddns'), 'domains', str(id)])
        if records is True:
            url = query(url, showRecords='true')
        if subdomains is True:
            url = query(url, showSubdomains='true')
        return handle_request('get', url, wrapper=cls)

    @classmethod
    def create(cls, name, ttl=300, records=None, subdomains=None, comment=None,
               email_address=None):
        """Create a CloudDNS Domain.

        :param name: A domain name such as ``yourname.com``
        :type name: str
        :param ttl: Time-To-Live (TTL) in seconds
        :type ttl: int
        :param records: A list of Records to create
        :type records: list of :class:`Record`
        :param subdomains: A list of Subdomains to create
        :type subdomains: list of :class:`Subdomain`
        :param comment: An optional comment to associated with the domain
        :type comment: str
        :param email_address: An e-mail address to associated with the domain
        :type email_address: str
        :returns: A shiny new Domain.
        :rtype: :class:`Domain`

        .. versionadded:: 0.1
        """
        data = {'domains': [{'name': name,
                             'ttl': int(ttl),
                             'recordsList': {'records': []},
                             'subDomains': {'domains': []}}]}
        create_ns_records = True
        if isinstance(records, (list, tuple)):
            for record in records:
                if isinstance(record, Record):
                    if record.type == 'NS':
                        create_ns_records = False
                    data['domains'][0]['recordsList']['records'].append({
                            'name': record.name,
                            'type': record.type,
                            'data': record.data,
                            'ttl': record.ttl,
                            'priority': record.priority,
                            'comment': record.comment
                            })
        if create_ns_records:
                data['domains'][0]['recordsList']['records'].append({
                        'name': name,
                        'type': 'NS',
                        'data': 'dns1.stabletransit.com',
                        'ttl': 3600
                        })
                data['domains'][0]['recordsList']['records'].append({
                        'name': name,
                        'type': 'NS',
                        'data': 'dns2.stabletransit.com',
                        'ttl': 3600
                        })
        if isinstance(subdomains, (list, tuple)):
            for subdomain in subdomains:
                if isinstance(subdomain, Subdomain):
                    data['domains'][0]['subdomains']['domains'].append({
                            'name': subdomain.name,
                            'emailAddress': subdomain.email_address,
                            'comment': subdomain.comment
                            })
        if comment is not None:
            data['domains'][0]['comment'] = comment
        if email_address is not None:
            data['domains'][0]['email_address'] = email_address
        data = json.dumps(data)
        url = '/'.join([get_url('clouddns'), 'domains'])
        return handle_request('post', url, data, cls, 'domains')

    @classmethod
    def import_zone(cls, contents, type='BIND_9'):
        """Import a raw BIND zone into CloudDNS.

        :param contents: Contents of the BIND zone
        :type contents: str
        :param type: The BIND format type being used, such as ``BIND_9``
        :type type: str
        :returns: A list of :class:`Domain`

        .. versionadded:: 0.1
        """
        data = {'domains': [{'contentType': type,
                             'contents': contents}]}
        data = json.dumps(data)
        url = '/'.join([get_url('clouddns'), 'import'])
        return handle_request('post', url, data, cls, 'domains')


class Export(DotDict):
    """A CloudDNS BIND Zone Export."""
    pass


class Nameserver(DotDict):
    """A CloudDNS Nameserver."""
    def __repr__(self):
        if 'name' in self:
            return '<Nameserver %s>' % self['name']
        return super(Nameserver, self).__repr__()


class Record(DotDict):
    """A CloudDNS Record."""
    def __repr__(self):
        if 'name' in self:
            return '<Record %s>' % self['name']
        return super(Record, self).__repr__()

    def __setitem__(self, key, value):
        if key in ['created', 'updated']:
            value = convert_datetime(value)
        super(Record, self).__setitem__(key, value)

    @classmethod
    def create(cls, name, type, data, ttl=None, priority=None, comment=None):
        """Create a CloudDNS Record.

        .. important::

            This is only a factory method. In order for the Record to be
            created on CloudDNS you must add the :class:`Record` this method
            returns to an existing :class:`Domain`.

        :param name: A domain name such as ``www.mydomain.com``
        :type name: str
        :param type: A record type, such as ``A``, ``CNAME``, ``MX``, etc.
        :type type: str
        :param data: Data associated with the record (depends on `type`)
        :type data: str
        :param ttl: Time-To-Live (TTL) in seconds
        :type ttl: int
        :param priority: Weighted priority (required for ``MX`` records)
        :type priority: int
        :param comment: An optional comment or note for this record
        :type comment: str
        :returns: A :class:`Record`

        .. versionadded:: 0.1
        """
        if ttl is not None:
            ttl = int(ttl)
        if priority is not None:
            priority = int(priority)
        return cls(name=name, type=type, data=data, ttl=ttl, priority=priority,
                   comment=comment)

    def reload(self):
        """Reload a Record.

        :returns: A Record
        :rtype: :class:`Record`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        assert 'domain_id' in self
        url = '/'.join([get_url('clouddns'), 'domains', str(self['domain_id']),
                                'records', str(self['id'])])
        response = handle_request('get', url, wrapper=Record)
        self.update(response)
        return self

    def modify(self, name=None, data=None, ttl=None):
        """Modify this Record's properties.

        :param name: Modify the Record's name.
        :type ttl: str
        :param data: Modify the Record's data.
        :type data: str
        :param ttl: Modify the Record's time-to-live (TTL).
        :type ttl: int
        :returns: A list of Records.
        :rtype: A list of :class:`Record`

        .. versionadded:: 0.1
        """
        assert 'id' in self
        assert 'domain_id' in self
        data = {}
        if name is not None:
            data['name'] = name
        if data is not None:
            data['data'] = data
        if ttl is not None:
            data['ttl'] = int(ttl)
        data = json.dumps(data)
        url = '/'.join([get_url('clouddns'), 'domains', str(self['domain_id']),
                        'records', str(self['id'])])
        handle_request('put', url, data)
        if name is not None:
            self['name'] = name
        if data is not None:
            self['data'] = data
        if ttl is not None:
            self['ttl'] = int(ttl)
        return self

    def delete(self, subdomains=False):
        """Delete this Record.

        .. warning::

            There is no confirmation step to this operation. Deleting this
            record is permanent. If in doubt you can export a copy of the DNS
            zone (:func:`vaporize.domains.Domain.export_zone`) before deleting.

        .. versionadded:: 0.1
        """
        assert 'id' in self
        assert 'domain_id' in self
        url = '/'.join([get_url('clouddns'), 'domains', str(self['domain_id']),
                        'records', str(self['id'])])
        handle_request('delete', url)


class Subdomain(DotDict):
    """A CloudDNS Subdomain."""
    def __repr__(self):
        if 'name' in self:
            return '<Subdomain %s>' % self['name']
        return super(Nameserver, self).__repr__()

    @classmethod
    def create(cls, name, comment=None, email_address=None):
        """Create a Subdomain.

        .. important::

            This is only a factory method. In order for the Subdomain to be
            created the :class:`Subdomain` returned by this method must be
            added to an existing CloudDNS :class:`Domain`.

        :param name: A subdomain such as ``www.yourdomain.com``
        :type name: str
        :param comment: An optional comment associated with the subdomain
        :type comment: str
        :param email_address: An e-mail address associated with the subdomain
        :type email_address: str
        :returns: A Subdomain
        :rtype: :class:`Subdomain`

        .. versionadded:: 0.1
        """
        return cls(name=name, comment=comment, email_address=None)



# -*- coding: utf-8 -*-

import json


class ConnectionError(Exception):
    pass


class BadRequest(Exception):
    pass


class Unauthorized(Exception):
    pass


class Forbidden(Exception):
    pass


class NotFound(Exception):
    pass


class InternalServerError(Exception):
    pass

class ServiceUnavailable(Exception):
    pass


class OverLimit(Exception):
    pass


class BuildInProgress(Exception):
    pass


class BadMediaType(Exception):
    pass


class UnknownError(Exception):
    pass


def handle_exception(code, msg):
    if code == 400:
        raise BadRequest(msg)
    elif code == 401:
        raise Unauthorized(msg)
    elif code == 403:
        raise Forbidden(msg)
    elif code == 404:
        raise NotFound(msg)
    elif code == 409:
        raise BuildInProgress(msg)
    elif code == 413:
        raise OverLimit(msg)
    elif code == 415:
        raise BadMediaType(msg)
    elif code == 500:
        raise InternalServerError(msg)
    elif code == 503:
        raise ServiceUnavailable(msg)
    else:
        raise UnknownError(msg)
